/* $Id: version.h,v 1.77.2.15 1998/06/25 17:23:38 wessels Exp $
 *
 *  SQUID_VERSION - String for version id of this distribution
 */
#ifndef SQUID_VERSION
#define SQUID_VERSION	"1.1.22"
#endif

#ifndef SQUID_RELEASE_TIME
#define SQUID_RELEASE_TIME 898795625
#endif
