
/*
 * $Id: icmp.h,v 1.8 1996/11/06 23:14:40 wessels Exp $
 *
 * DEBUG: section 33    Client-side Routines
 * AUTHOR: Duane Wessels
 *
 * SQUID Internet Object Cache  http://squid.nlanr.net/Squid/
 * --------------------------------------------------------
 *
 *  Squid is the result of efforts by numerous individuals from the
 *  Internet community.  Development is led by Duane Wessels of the
 *  National Laboratory for Applied Network Research and funded by
 *  the National Science Foundation.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  
 */

#ifndef _ICMP_H_
#define _ICMP_H_

extern int icmp_sock;

extern void icmpOpen _PARAMS((void));
extern void icmpClose _PARAMS((void));
extern void icmpPing _PARAMS((struct in_addr to));
extern void icmpSourcePing _PARAMS((struct in_addr to, const icp_common_t *, const char *url));
extern void icmpDomainPing _PARAMS((struct in_addr to, const char *domain));

#endif /* _ICMP_H_ */
